#include <cstdio>
#include <cmath>
#include <map>

using namespace std;

#include "defs.hh"
#include "reader.hh"
#include "s3m.hh"
#include "freq2.hh"

static map<unsigned, unsigned> SmpID_to_InsNum;
static map<unsigned, unsigned> InsNum_to_SmpID;

static map<unsigned, double> wavelenmap;

static void WriteSamples(S3M &song)
{
	unsigned insnum = 0;
	map<unsigned, Sample>::const_iterator i;
	for(i=Samples.begin(); i!=Samples.end(); ++i)
	{
		S3M::instrument& ins = song.getinstrument(insnum);
		
		unsigned smpid = i->first;
		
		char Buf[64];
		sprintf(Buf, "Sample %X", i->first);
		
		ins.MakeSMP();
		ins.name = Buf;
		
		// Convert the audio from signed 16bit to unsigned 8-bit
		const string &srcaudio = i->second.data;
		vector<unsigned char> &destaudio = ins.audio;
		destaudio.resize(srcaudio.size() / 2);
		for(unsigned a=0; a<destaudio.size(); ++a)
		{
			unsigned char lo = srcaudio[a*2+0];
			unsigned char hi = srcaudio[a*2+1];
			signed short value = (lo + (hi << 8));
			destaudio[a] = 0x80 + (value / 256);
		}
		
		if(i->second.looping)
			ins.SetLoop(i->second.loopbeg, i->second.loopend);

		double wavelen = GetWaveLen(smpid);
		
		/*  Information bits from S3M documentation:
		 * 
		 * C4SPD is the frequency of the note C4.
		 *
		 * C4 is 261.6255 Hz.
		 * 
		 * Thus if C4SPD is 8363, it means the wave
		 * length is 8363Hz / 261.2655Hz = 32.
		 * 
		 * wavelen = C4Spd / C4freq
		 * 
		 * C4Spd = wavelen * C4freq
		 */
		
		const double C4freq = 261.6255653;
		
		unsigned C4Spd = (unsigned)(wavelen * C4freq);
		while(C4Spd > 50000) C4Spd /= 2;
		while(C4Spd < 1000) C4Spd *= 2;

		ins.SetC4SPD(C4Spd);
		
		/* C4Spd was rounded and beated.
		 * Take wavelen back!
		 */
		wavelen = C4Spd / C4freq;
		wavelenmap[smpid] = wavelen;

		fprintf(stderr, "Sample %u C4Spd therefore %u. Wavelen updated to %g.\n",
			smpid, C4Spd, wavelen);

		SmpID_to_InsNum[smpid] = insnum;
		InsNum_to_SmpID[insnum] = smpid;
		
		++insnum;
	}
}

static unsigned char getnote(unsigned long freq, unsigned insno)
{
	unsigned smpid = Instrs[insno].sampleid;
	double basefreq = 44100.0 / wavelenmap[smpid];

	// see spctomidi.cc for comments what this means
	double n = freq / 65536.0;
	double note = 17.31234049066756 * log(n * basefreq) - 36.37631656229592;
	
	note -= 12;
	
	if(note<1)note = 1; if(note>90)note=90;
	unsigned nn = (unsigned) note;
	
	return (nn / 12) * 16 + (nn % 12);
}

#define ROWTIME 0.10
static void BuildChannel(const vector<Event>& spcevents, vector<S3M::event>& s3mevents)
{
	double s3mtime=0;
	
	unsigned lastline=0;
	unsigned freq=0;
	unsigned note=0;
	double lastfreq=0;
	
	int maxvol=0;
	bool noteon=false;
    for(unsigned e=0; e<spcevents.size(); ++e)
    {
        const Event &ev = spcevents[e];
        if(ev.evtype=='O' && !ev.ins) continue; // instrument unknown, thus ignore!
        
        double evtime = ev.time;

		double timeused = 0;
        unsigned newrowcount = 0;
        while(s3mtime+timeused < evtime)
        {
        	newrowcount++;
        	timeused += ROWTIME;
        }
        
        if(!s3mevents.size() && !newrowcount)newrowcount = 1;
        if(ev.evtype=='O' && !newrowcount)newrowcount = 1;
        
        s3mtime += newrowcount * ROWTIME;

       	s3mevents.resize(s3mevents.size() + newrowcount);
       	for(unsigned r=0; r<newrowcount; ++r)
       	{
       		// handle the lines
       	}
       	
		//note(255=no),ins(0=no),vol(255=no),cmd(0=no,1=A),info(0=def)
        S3M::event &thisline = s3mevents[s3mevents.size() - 1];
        
        switch(ev.evtype)
        {
        	case 'K': // keyoff
        		//thisline.vol = 0;
        		break;
        	case 'O': // keyon (.ins, use last known freq)
        		thisline.note = getnote(freq, ev.ins);
        		thisline.ins  = SmpID_to_InsNum[Instrs[ev.ins].sampleid] + 1;
        		thisline.vol  = maxvol;
        		lastfreq = freq;
        		noteon = true;
        		break;
        	case 'F': // freq (.freq)
        		freq = ev.freq;
        		break;
        	case 'V': // volumes (.l, .r)
        		int lvol = ev.l;
        		int rvol = ev.r;
        		maxvol = lvol >? rvol;
        		if(maxvol > 64)maxvol = 64;
        		if(noteon)
        		    thisline.vol = maxvol;
        		break;
        }
    }
}

static void BuildChannels(S3M &song)
{
	vector<S3M::event> s3mevents[8];
	unsigned maxrows=0;
	for(unsigned a=0; a<8; ++a)
	{
		BuildChannel(Channels[a].events, s3mevents[a]);
		
		for(unsigned l=0; l<s3mevents[a].size(); ++l)
		{
			const S3M::event &e = s3mevents[a][l];
        	static const char notes[12*3+1]=
        	"C-\0C#\0D-\0D#\0E-\0F-\0F#\0G-\0G#\0A-\0A#\0B-\0";
			fprintf(stderr, "[%u]: ", a);
			if(e.note == 255) fprintf(stderr, "...");
			else if(e.note == 254) fprintf(stderr, "^^^");
			else fprintf(stderr, "%s%d", notes+3*(e.note%12), e.note/12);
			if(e.ins == 0)fprintf(stderr, " ..");
			else fprintf(stderr, " %02u", e.ins);
			if(e.vol == 255)fprintf(stderr, " ..");
			else fprintf(stderr, " %02u", e.vol);
			if(e.cmd == 0)fprintf(stderr, " .");
			else fprintf(stderr, " %c", e.cmd+64);
			fprintf(stderr, "%02X - %g\n", e.info, l*ROWTIME);
		}
		
		if(s3mevents[a].size() > maxrows)
			maxrows = s3mevents[a].size();
	}
	
	/* Simply fill the patterns with data we just got */
	
	S3M::eventtable e;
	e.setgeometry(maxrows, 8);
	for(unsigned a=0; a<8; ++a)
		e.loadchannel(a, s3mevents[a]);

	unsigned patcount = (maxrows+63) / 64;
	for(unsigned p=0; p<patcount; ++p)
	{
		if(p == 100) break;
		S3M::pattern pat = song.compresspattern(e.slice(p*64, 64));
		
		song.getpattern(p) = pat;
		song.setorder(p, p);
	}
}

static void WriteS3M(void)
{
	S3M song;
	
	for(unsigned c=0;c<8;++c)
	{
		song.setupchannel(c*2,   S3M::leftchan(c));
		song.setupchannel(c*2+1, S3M::rightchan(c));
	}
	
	WriteSamples(song);
	BuildChannels(song);
	
	song.setsettings(3, 0xA0,
	                 63,63, false);
	
	song.Save("s3mfile.s3m");
	song.Dump();
}

int main(void)
{
    ReadLog(stdin);
    
    WriteS3M();

    return 0;
}
