#include <cstdio>

#include "core.hh"

#ifdef __sun
#ifdef __SVR4
#include <sys/audioio.h>
#else
#include <sun/audioio.h>
#endif
#endif

#ifdef linux
#include <sys/soundcard.h>
#endif

#include <sys/ioctl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>

#include <string>

#include <termio.h>

using namespace std;

/* LOGMUL = exp(log(2) / 3) */
#define LOGMUL 1.259921049894873

static struct termio back;

void Exit ()
{
    if(ioctl(0, TCSETA, &back) < 0)
        fcntl(0, F_SETFL, fcntl(0, F_GETFL) &~ O_NONBLOCK);

    SPCCore_MuteSound();
    exit(0);
}

#include "argh.hh"

#define VERSIONSTR \
    "spcplay - SPC player based on snes9x\n" \
    "snes9x copyright http://www.snes9x.com/,\n" \
    "player copyright (C) 2001,2003 Bisqwit (http://iki.fi/bisqwit/)\n" \
    "Thanks for the xmms spc plugin authors.\n"

static string filee;
#ifdef linux
static string snd_dev = "/dev/dsp";
#endif
#ifdef __sun
static string snd_dev = "/dev/audio";
#endif
static int surround = 0;
static int reversestereo = 0;

class Handle : public arghandler
{
    bool Help;
private:
    string opt_h(const string &s)
    {
        Help = true;
        return s;
    }
    string opt_V(const string &)
    {
        printf(VERSIONSTR);
        exit(0);
    }
    string opt_2(const string &s) { SPCCore_SetChannels(2);                   return s; }
    string opt_m(const string &s) { SPCCore_SetChannels(1);                   return s; }
    string opt_b(const string &s) { SPCCore_SetBitness(8);                 return s; }
    string opt_6(const string &s) { SPCCore_SetBitness(16);                return s; }
    string opt_n(const string &s) { SPCCore_SetInterpolation(0);           return s; }
    string opt_i(const string &s) { SPCCore_SetInterpolation(1);           return s; }
    string opt_e(const string &s) { SPCCore_SetEcho(1);                    return s; }
    string opt_c(const string &s) { SPCCore_SetEcho(0);                    return s; }
    string opt_a(const string &s) { SPCCore_SetAlt(1);    return s; }
    string opt_u(const string &s) { surround = 1;                       return s; }
    string opt_s(const string &s) { reversestereo = 1;                  return s; }
    string opt_r(const string &s)
    {
        const char *q = s.c_str();
        const char *p = q;
        long rate = strtol(p, (char **)&p, 10);
        if(rate < 18 || rate > 256000)
        {
            argerror(s);
        }
        else
            SPCCore_SetRate(rate);
        return s.substr(p-q);
    }
    string opt_o(const string &s)
    {
        snd_dev = s;
        return "";
    }
public:
    Handle(const char *defopts, int argc, const char *const *argv)
    : arghandler(defopts, argc, argv),
        Help(false)
    {
        add("-2", "--stereo", "Specifies stereo sound. Redundant. Default.", &Handle::opt_2);
        add("-s", "--reverse","Reverse stereo.",                             &Handle::opt_s);
        add("-m", "--mono",   "Specifies mono sound.",                       &Handle::opt_m);
        add("-b", "--byte",   "8-bit mode. Normally 16-bit mode.",           &Handle::opt_b);
        add("-6", "--word",   "16-bit mode. Redundant.",                     &Handle::opt_6);
        add("-r", "--rate",   "Specifies playing rate. 44410 is default.",   &Handle::opt_r);
        add("-i", "--ipolate","Enables interpolating. Default.",             &Handle::opt_i);
        add("-n", "--npolate","Disables interpolating. (Toggle-key: 9/i)",   &Handle::opt_n);
        add("-e", "--echo",   "Enables echo. Default.",                      &Handle::opt_e);
        add("-c", "--noecho", "Disables echo. (Toggle-key: e)",              &Handle::opt_c);
        add("-o", "--output", "Specify output device (default: /dev/dsp).",  &Handle::opt_o);
        add("-h", "--help",   "Help",                                        &Handle::opt_h);
        add("-V", "--version","Version information",                         &Handle::opt_V);
        add("-a", "--altsampledecode",
                               "Better sample decoding (Toggle-key: a)",         &Handle::opt_a);
        add("-u", "--surround","Flip one channel (stereo only) (Toggle-key: u)", &Handle::opt_u);
    }
    virtual void defarg(const string &arg)
    {
        if(arg.size())filee = arg;
    }
    void Usage()
    {
        fprintf(stderr, VERSIONSTR);
        fprintf(stderr,
            "Usage: %s [<options> ...] <file>\n"
            "\n"
            "Emulates and plays SPC files (snes soundtracks).\n"
            "\n"
            "Options:\n", a0.c_str());
        
        listoptions();
        
        exit(0);
    }
    virtual void parse()
    {
        arghandler::parse();
        if(Help || !filee.size())Usage();
    }
};

static double mastervolume = 1;
static double extrastereo  = 1;

static int cur_rate;

static void State(void)
{
     fprintf(stderr, "\rRate %-4d", cur_rate);
     fprintf(stderr, "   Volume %-4.0f ", (mastervolume*100+0.5));
     if(extrastereo != 1.0)
     {
         fprintf(stderr, "   Stereo %-4.0f%% ", (extrastereo*100+0.5));
     }
     for(unsigned a=0; a<8; ++a)
         fputc(SPCCore_GetChannelStatus(a) ? (a+'1') : '_', stderr);
     fprintf(stderr, "%20s", "");
     fflush(stderr);
}

static void KeyInit()
{
    struct termio term = back;
    
    // Disable linebuffer and echoing
    term.c_lflag &= ~(ICANON|ECHO);
    // Disable signals generated by ^C, ^Z and ^] (int,stop,quit)
    //term.c_lflag &= ~(ISIG);
    
    // Disable special handling of ^S and ^Q
    //term.c_iflag &= ~(IXON);
    
    term.c_cc[VMIN] = 0; // 0=no block, 1=do block
    if(ioctl(0, TCSETA, &term) < 0)
        fcntl(0, F_SETFL, fcntl(0, F_GETFL) | O_NONBLOCK);
}

static void Resume(int)
{
    KeyInit();
}

int main(int argc, const char *const *argv)
{
    Handle parameters("", argc, argv);
    parameters.parse();
    
    ioctl(0, TCGETA, &back);
    signal(SIGCONT, Resume);
    KeyInit();
    
    unsigned bufsize = SPCCore_Init();
    unsigned char *Buf = (unsigned char *)malloc(bufsize);
    
#if 0
    for(unsigned fni=0; fni<files.size(); ++fni)
#else
    for(unsigned i=0; i<1; ++i)
#endif
    {
#if 0
        const char *fn = files[fni].c_str();
#else
        const char *fn = filee.c_str();
#endif
        if(SPCCore_Load(fn))
        {
            string s = fn;
            s += " could not be opened";
            perror(s.c_str());
            break;
        }
        
        int fd;
        
        if(snd_dev=="-" || !snd_dev.size())
            fd = STDOUT_FILENO;
        else
        {
#ifdef __sun
            fd = open(snd_dev.c_str(), O_WRONLY);
            if(fd >= 0)
            {
                audio_info_t audio;
                AUDIO_INITINFO(&audio);
                audio.play.sample_rate = SPCCore_GetRate();
                audio.play.channels = SPCCore_GetChannels();
                audio.play.precision = SPCCore_GetBitness();
                audio.play.encoding = AUDIO_ENCODING_LINEAR;
                ioctl(fd, AUDIO_SETINFO, &audio);
            }
#endif
#ifdef linux
            fd = open(snd_dev.c_str(), O_WRONLY);
            if(fd >= 0)
            {
                int stereo = SPCCore_GetChannels()-1;
                int res = SPCCore_GetBitness();
                int rate = SPCCore_GetRate();
                ioctl(fd, SNDCTL_DSP_SAMPLESIZE, &res);
                ioctl(fd, SNDCTL_DSP_STEREO,     &stereo);
                ioctl(fd, SNDCTL_DSP_SPEED,      &rate);
                SPCCore_SetBitness(res);
                SPCCore_SetRate(rate);
                SPCCore_SetChannels(stereo?2:1);
            }
#endif
        }
        
        if(fd < 0)
        {
            string s = snd_dev + " could not be opened";
            fprintf(stderr, "%s\n", s.c_str());
            break;
        }
        
        //memset(Buf, 0, bufsize);
        //write(fd, Buf, bufsize);
        
        fprintf(stderr, "File: %s\n", fn);
        
        SPCCore_SetPlaybackSpeed(cur_rate = 100);
        
        SPCCore_ID *id = SPCCore_GetID(fn);
        if(id)
        {
            char Unkn[64];sprintf(Unkn,"Unknown (%d)", (int)id->emulator);
            fprintf(stderr,
                 "Song: %-32s     Game: %-32s\n"
                 "Comments: %-32s Author: %-32s\n"
                 "Dumper: %-16s   Playtime: %d   Fadetime: %d   Emulator: %s\n",
                 id->songname,
                 id->gametitle,
                 id->comments,
                 id->author,
                 id->dumper,
                 id->playtime,
                 id->fadetime,
                 (id->emulator==1) ? "ZSnes"
               : (id->emulator==2) ? "snes9x"
               : Unkn);
        }
        else
            fprintf(stderr, "(no id666 tag)\n");
        
        State();
        nice(-40);
        
        fcntl(1, F_SETFL, fcntl(1, F_GETFL) & ~O_NONBLOCK);
        for(;;)
        {
            unsigned char c;
            while(read(0, &c, 1) == 1)
            {
                if(c=='q')goto Done;
                if(c=='u')surround ^= 1;
                if(c=='s')reversestereo ^= 1;
                if(c=='a')SPCCore_SetAlt(!SPCCore_GetAlt());
                if(c=='9' || c=='i')
                    SPCCore_SetInterpolation(
                      (SPCCore_GetInterpolation()+1) % SPCCore_GetNumInterpolation()
                                            );
                if(c=='e')SPCCore_SetEcho(!SPCCore_GetEcho());
                if(c=='c')
                {
                    if(cur_rate<50)++cur_rate;else cur_rate+=10;
                    SPCCore_SetPlaybackSpeed(cur_rate);
                    cur_rate = SPCCore_GetPlaybackSpeed();
                }
                if(c=='d')
                {
                    if(cur_rate>50)cur_rate-=10;else
                    if(cur_rate>1)--cur_rate;
                    SPCCore_SetPlaybackSpeed(cur_rate);
                    cur_rate = SPCCore_GetPlaybackSpeed();
                }
                if(c=='+')
                    mastervolume *= LOGMUL;
                if(c=='-')
                {
                    mastervolume /= LOGMUL;
                    if(mastervolume < 0.004)mastervolume = 0.004;
                }
                if(c=='w')
                    extrastereo *= LOGMUL;
                if(c=='W')
                {
                    extrastereo /= LOGMUL;
                    if(extrastereo < 0.00984)mastervolume = 0.00984;
                }
                if(c>='1' && c<='8')
                {
                    SPCCore_SetChannelStatus(c-'1',
                   !SPCCore_GetChannelStatus(c-'1'));
                }
                else if(c=='0')
                {
                    for(unsigned n=0; n<8; ++n)
                        SPCCore_SetChannelStatus(n, 1);
                }
                State();
            }
            
            SPCCore_Generate(Buf);
            
            /* LITTLE-ENDIAN! */
            if(SPCCore_GetBitness() == 16)
            {
                signed short *smp = (signed short *)Buf;
                const unsigned count = bufsize / sizeof(*smp);
                for(unsigned a=0; a<count; ++a)
                {
                    int v = static_cast<int> (*smp * mastervolume);
                    if(surround && (a&1) && SPCCore_GetChannels() > 1)v=-v;
                    if(v > 32767)v = 32767;else if(v < -32768)v = -32768;
                    *smp++ = v;
                }
                if(reversestereo && SPCCore_GetChannels() == 2)
                {
                    smp = (signed short *)Buf;
                    for(unsigned a=0; a<count; ++a)
                    {
                        int v1 = smp[0];
                        int v2 = smp[1];
                        smp[0] = v2;
                        smp[1] = v1;
                        smp += 2;
                        ++a;
                    }
                }
                if(extrastereo != 1.0 && SPCCore_GetChannels() == 2)
                {
                    smp = (signed short *)Buf;
                    for(unsigned a=0; a<count; ++a)
                    {
                        int v1 = smp[0];
                        int v2 = smp[1];
                        
                        double avg = (v1 + v2) / 2.0;
                        int l = avg + extrastereo * (v1 - avg);
                        int r = avg + extrastereo * (v2 - avg);
                        if(l > 32767)l = 32767;else if(l < -32768)l = -32768;
                        if(r > 32767)r = 32767;else if(r < -32768)r = -32768;
                        
                        smp[0] = (int)l;
                        smp[1] = (int)r;
                        smp += 2;
                        ++a;
                    }
                }
            }
            else
            {
                /* FIXME: Volume is ignored for 8-bit */
            }
            
            write(fd, Buf, bufsize);
        }
Done:
        close(fd);
    }
    
    free(Buf);
    
    SPCCore_Terminate();
    
    Exit();
    return 0;
}
